@@ -14,7 +14,8 @@ import asyncio
 
 class YFClient:
 
-    BASE_URL = 'https://query1.finance.yahoo.com'# {{{
+    snake_pattern = re.compile(r'(?<!^)(?=[A-Z])')# {{{
+    BASE_URL = 'https://query1.finance.yahoo.com'
     SCREENER_URL = BASE_URL + '/v1/finance/screener'
     QUOTE_SUMMARY_URL = BASE_URL + '/v10/finance/quoteSummary/'
     TRENDING_URL = BASE_URL + '/v1/finance/trending/'
@@ -54,8 +55,15 @@ class YFClient:
             'preMarketVolume', 'preMarketOpen', 'currentTradingPeriod', 'tradingPeriods',
             'preMarketPrice', 'preMarketDayHigh', 'preMarketDayLow', 'preMarketPreviousClose',
             'regularMarketPreviousClose', 'postMarketPreviousClose', 'postMarketVolume',
-            'marketState', 'postMarketTime', 'regularMarketClose', 'last', 'trade',
-            'postMarketChangePercent', 'postMarketPrice', 'postMarketChange'] #not used rn
+            'postMarketTime', 'regularMarketClose', 'last', 'trade',
+            'postMarketChangePercent', 'postMarketPrice', 'postMarketChange',
+            'language', 'region', 'quoteType', 'quoteSourceName', 'triggerable',
+            'sourceInterval', 'exchangeDataDelayedBy', 'tradeable', 'firstTradeDateMilliseconds',
+            'priceHint', 'regularMarketTime', 'regularMarketDayHigh', 'regularMarketDayRange',
+            'regularMarketDayLow', 'fullExchangeName', 'fiftyTwoWeekLowChange', 'fiftyTwoWeekLowChangePercent',
+            'fiftyTwoWeekRange', 'fiftyTwoWeekHighChange', 'fiftyTwoWeekHighChangePercent',
+            'exchange', 'exchangeTimezoneName', 'exchangeTimezoneShortName', 'gmtOffSetMilliseconds',
+            'market', 'esgPopulated', 'sharesOutstanding'] #not used rn
 
 
     SCREENER_COLS = ['symbol', 'exchange', 'currency', 'fullExchangeName', 'quoteSourceName', 
@@ -63,6 +71,9 @@ class YFClient:
             'priceHint', 'market', 'messageBoardId', 'financialCurrency', 'sourceInterval', 'exchangeDataDelayedBy',
             'exchangeTimezoneName', 'exchangeTimezoneShortName', 'gmtOffSetMilliseconds', 'prevName', 'nameChangeDate']
 
+    PRICE_HISTORY_COLS = ['symbol', 'date', 'close', 'volume', 'low', 'high',
+                'adjclose', 'ratio', 'adjopen', 'adjhigh', 'adjlow', 'interval', 'date_local',
+                'trading_period']
 
 # }}}
 
@@ -84,6 +95,14 @@ class YFClient:
 
         # }}}
 
+    async def get_new_crumb(self):# {{{
+        self.crumb, self.cookies = self._get_crumb()
+        await self.session.close()
+        connector = aiohttp.TCPConnector(limit=10)
+        self.session = aiohttp.ClientSession(connector=connector,
+                headers=self.HEADERS, cookies=self.cookies)
+        # }}}
+
     def _get_crumb(self):# {{{
         """method to retrieve the crumb that yahoo uses for api authentication"""
 
@@ -109,6 +128,193 @@ class YFClient:
     async def disconnect(self):# {{{
         await self.session.close()# }}}
 
+    async def _get_quote(self, symbols):# {{{
+
+        url = self.QUOTE_URL
+
+        resp = await self._make_request('get', url, params={'symbols': ','.join(symbols), 'fields': ','.join(self.QUOTE_COLUMNS)})
+
+        resp = await resp.json()
+        if resp['quoteResponse']['error']:
+            raise ValueError(f'{resp["quoteResponse"]["error"]}')
+
+        res = resp['quoteResponse']['result']
+        if res:
+            return res
+        else:
+            return [{'symbol': x} for x in symbols]
+        # }}}
+
+    async def get_quote(self, symbols, columns=None):# {{{
+
+        if isinstance(columns, str):
+            columns = [columns]
+
+        if isinstance(symbols, str):
+            if ',' in symbols:
+                symbols = symbols.split(',')
+            else:
+                symbols = [symbols]
+
+
+        chunks = [symbols[x:x+50] for x in range(0, len(symbols), 50)]
+        tasks = []
+        for chunk in chunks:
+            tasks.append(self._get_quote(chunk))
+
+        res = await asyncio.gather(*tasks)
+
+        empty_df = pd.DataFrame(columns=self.QUOTE_COLUMNS)
+
+        df = pd.concat([empty_df] + [pd.DataFrame(x) for x in res])
+
+        if columns:
+            if 'symbol' not in columns:
+                columns += ['symbol']
+            df = df[columns]
+
+        for c in df.columns:
+            if c not in self.QUOTE_COLUMNS:
+                print(f'column {c} found in quote df but not in self.QUOTE_COLUMNS')
+
+        for c in self.QUOTE_COLUMNS:
+            if c not in df.columns:
+                df[c] = None
+
+        return df
+        # df.columns = self.camel_to_snake(list(df.columns))
+
+        # return df# }}}
+
+    async def _price_history(self, symbol, interval=None, start=None, end=None, adjust=True,# {{{
+            prepost=False):
+
+        print(f'getting price history for {symbol}')
+
+        assert isinstance(end, datetime.datetime); assert end.tzinfo
+        assert isinstance(start, datetime.datetime); assert start.tzinfo
+
+        url = self.HISTORY_URL.format(symbol=symbol)
+
+
+        params = {'includeAdjustedClose': str(adjust).lower(), 'events': 'div,splits,capitalGain',
+                'interval': interval, 'includePrePost': str(prepost).lower()}
+
+
+
+        # put the start and end in (micro) seconds
+        params['period1'] = int(time.mktime(start.timetuple()))
+        params['period2'] = int(time.mktime(end.timetuple()))
+
+
+
+        resp = await self._make_request('get', url, params=params)
+
+        resp = await resp.json()
+        if 'finance' in resp.keys():
+            print('finance error')
+            print(f'{resp["finance"]["error"]}')
+            return
+        if resp['chart']['error']:
+            if resp['chart']['error']['description'][0:35] == 'Data doesn\'t exist for startDate = ':
+                print("no results found for period")
+                return
+            else:
+                print('chart error')
+                print(f'{resp["chart"]["error"]}')
+                return
+
+        if not resp['chart']['result'][0]['indicators']['quote'][0]:
+            print(f"no results found for period")
+            return
+
+        meta = resp['chart']['result'][0]['meta']
+        resp = resp['chart']['result'][0]
+
+
+        timestamps = pd.to_datetime(resp['timestamp'], unit='s', utc=True)
+
+        data = pd.DataFrame(resp['indicators']['quote'][0], index=timestamps)
+
+        # adjusting other columns too
+        if adjust and 'adjclose' in resp['indicators'].keys():
+            data['adjclose'] = resp['indicators']['adjclose'][0]['adjclose']
+            data['ratio'] = data['adjclose'] / data['close']
+            data['adjopen'] = data['open'] * data['ratio']
+            data['adjhigh'] = data['high'] * data['ratio']
+            data['adjlow'] = data['low'] * data['ratio']
+        # elif adjust and 'adjclose' not in resp['indicators'].keys():
+        #     data['adjclose'] = data['close'].copy()
+        #     data['ratio'] = 1
+        #     data['adjopen'] = data['open'].copy()
+        #     data['adjhigh'] = data['high'].copy()
+        #     data['adjlow'] = data['low'].copy()
+
+
+        data = data.reset_index().rename(columns={'index': 'date'})
+        data['symbol'] = symbol
+        data['interval'] = interval.lower()
+
+        cols = list(data.columns)
+        cols.remove('symbol')
+        cols.remove('date')
+
+
+        data = data[['symbol', 'date'] + cols]
+
+        # now for pre and post items
+        thours = self._get_session_from_tperiod(meta)
+
+        data['date_local'] = data['date'].dt.tz_convert(thours['tz'])
+
+        data['trading_period'] = None
+
+        data.loc[(data['date_local'].dt.time>=thours['regular_start']) &
+                (data['date_local'].dt.time < thours['regular_end']), 'trading_period'] = 'regular'
+
+        data.loc[(data['date_local'].dt.time>=thours['pre_start']) &
+                (data['date_local'].dt.time < thours['pre_end']), 'trading_period'] = 'pre'
+
+        data.loc[(data['date_local'].dt.time>=thours['post_start']) &
+                (data['date_local'].dt.time< thours['post_end']), 'trading_period'] = 'post'
+
+        print(f'GOT for {symbol}')
+        return data# }}}
+
+    async def get_price_history(self, symbols, interval, start, end,# {{{
+            adjust=True, prepost=False):
+        """
+        interval: 1m, 2m, 5m, 15m, 30m, 60m, 90m, 1h, 1h, 1d, 5d, 1wk, 1mo, 3mo
+        adjust can only work for intraday
+        todo: workout what adjustment intraday actually does do (find a symbol that has recent one)
+        volume seems to only populate in normal market hours
+        start and end must be datetimes with timezones
+
+        if a ticker isn't found, it won't be returned in the data set
+        if no tickers are found; empty dataset is returned
+        """
+
+        if isinstance(symbols, str):
+            symbols = [symbols]
+
+        tasks = []
+        for t in symbols:
+            tmp = self._price_history(t, interval=interval, start=start,
+                    end=end, adjust=adjust, prepost=prepost)
+            tasks.append(tmp)
+        res = await asyncio.gather(*tasks)
+
+        fres = []
+        for r in res:
+            if r is not None:
+                fres.append(r)
+
+        if fres:
+            return pd.concat(fres)
+        else:
+            return pd.DataFrame(columns=self.PRICE_HISTORY_COLS)
+        # }}}
+
     def _build_screener_payload(self,# {{{
             sort_field='intradaymarketcap', sort_type='DESC', quote_type='EQUITY',
             mcap_filter=100_000_000, region='us'):
@@ -181,12 +387,18 @@ class YFClient:
         try:
             result = await resp.json()
         except:
-            cntnt = await resp.content
-            print(cntnt)
+            cntnt = await resp.text()
+            print(payload)
+            raise ValueError(f'couldnt convert to json {cntnt}')
 
         if result['finance']['error']:
             print(payload)
-            raise ValueError(f'{result["finance"]["error"]}')
+            if result['finance']['error']['description'] == 'Invalid Crumb':
+                await self.get_new_crumb()
+                print('getting new crumb')
+                return await self._send_screener_request(payload)
+            else:
+                raise ValueError(f'{result["finance"]["error"]}')
 
         # return pd.DataFrame(data)
         return result['finance']['result'][0]
@@ -228,23 +440,25 @@ class YFClient:
         payload = self._build_screener_payload(region=region, mcap_filter=mcap_filter)
         res = await self._iter_screener_requests(payload, max_results)
 
-        data = pd.DataFrame(res)
+        data = pd.DataFrame(columns=self.SCREENER_COLS)
+        data = data.append(pd.DataFrame(res))
 
-        cols = []
-        for c in self.SCREENER_COLS:
-            if c in data.columns:
-                cols.append(c)
-            else:
-                pass
-                # print(f'column {c} not found in ref data')
 
-        for c in data.columns:
-            if c not in self.SCREENER_COLS:
-                pass
-                # print(f'column {c} not found in defined screen columns')
+        # cols = []
+        # for c in self.SCREENER_COLS:
+        #     if c in data.columns:
+        #         cols.append(c)
+        #     else:
+        #         pass
+        #         # print(f'column {c} not found in ref data')
+
+        # for c in data.columns:
+        #     if c not in self.SCREENER_COLS:
+        #         pass
+        #         # print(f'column {c} not found in defined screen columns')
 
         print(f'GOT data')
-        return data[cols]# }}}
+        return data[self.SCREENER_COLS]# }}}
 
     async def get_all_equity_reference(self):# {{{
 
@@ -292,8 +506,13 @@ class YFClient:
         # res = self.session.get(url, params=params)
         res = await self._make_request('get', url, params=params)
 
-        res_js = (await res.json())['quoteSummary']['result'][0]
-        return res_js# }}}
+        res_js = await res.json()
+
+        if res_js['quoteSummary']['result']:
+            return res_js['quoteSummary']['result'][0]
+        else:
+            print(res_js['quoteSummary']['error'])
+            return {} # }}}
 
     async def get_quote_summary(self, symbols, modules=None):# {{{
 
@@ -332,8 +551,8 @@ class YFClient:
             print(f'got {region}')
             return tmp
         except Exception as e:
-            print(resp)
-            raise e
+            print(resp['finance'])
+            return []
         # }}}
 
     async def _get_screener_fields(self, asset_class):# {{{
@@ -366,147 +585,8 @@ class YFClient:
 
         return clean_regions# }}}
 
-    async def _get_quote(self, symbols):# {{{
-
-        print(f'getting data for {symbols}')
-
-        if isinstance(symbols, str):
-            symbols = [symbols]
-
-        url = self.QUOTE_URL
-
-        resp = await self._make_request('get', url, params={'symbols': ','.join(symbols), 'fields': ','.join(self.QUOTE_COLUMNS)})
-
-        resp = await resp.json()
-        if resp['quoteResponse']['error']:
-            raise ValueError(f'{resp["quoteResponse"]["error"]}')
-
-        print(f'GOT data')
-
-        return resp['quoteResponse']['result']# }}}
-
-    async def get_quote(self, symbols, columns=None):# {{{
-
-        if isinstance(columns, str):
-            columns = [columns]
-
-
-        chunks = [symbols[x:x+50] for x in range(0, len(symbols), 50)]
-        tasks = []
-        for chunk in chunks:
-            tasks.append(self._get_quote(chunk))
-
-        res = await asyncio.gather(*tasks)
-
-        df = pd.concat([pd.DataFrame(x) for x in res])
-
-        if columns:
-            if 'symbol' not in columns:
-                columns += ['symbol']
-            df = df[columns]
-
-        return df# }}}
-
-    async def _price_history(self, ticker, period=None, interval=None, start=None, end=None, adjust=True,# {{{
-            prepost=False):
-
-        print(f'getting price history for {ticker}')
-        url = self.HISTORY_URL.format(symbol=ticker)
-
-
-        params = {'includeAdjustedClose': str(adjust).lower(), 'events': 'div,splits,capitalGain',
-                'interval': interval, 'includePrePost': str(prepost).lower()}
-
-
-        assert isinstance(end, datetime.datetime)
-        assert end.tz
-        if period:
-            start, end = self._get_dates_from_period(period, end)
-        else:
-            assert isinstance(start, datetime.datetime)
-            assert start.tz
-
-        params['period1'] = int(time.mktime(start.timetuple()))
-        params['period2'] = int(time.mktime(end.timetuple()))
-
-
-
-        # print(params)
-        resp = await self._make_request('get', url, params=params)
-
-        resp = await resp.json()
-        if 'finance' in resp.keys():
-            raise ValueError(f'{resp["finance"]["error"]}')
-        if resp['chart']['error']:
-            raise ValueError(f'{resp["chart"]["error"]}')
-
-        if not resp['chart']['result'][0]['indicators']['quote'][0]:
-            raise ValueError(f"no results found for period")
-
-        meta = resp['chart']['result'][0]['meta']
-        resp = resp['chart']['result'][0]
-
-
-        timestamps = pd.to_datetime(resp['timestamp'], unit='s', utc=True)
-
-        data = pd.DataFrame(resp['indicators']['quote'][0], index=timestamps)
-
-        # adjusting other columns too
-        if adjust and 'adjclose' in resp['indicators'].keys():
-            data['adjclose'] = resp['indicators']['adjclose'][0]['adjclose']
-            data['ratio'] = data['adjclose'] / data['close']
-            data['adjopen'] = data['open'] * data['ratio']
-            data['adjhigh'] = data['high'] * data['ratio']
-            data['adjlow'] = data['low'] * data['ratio']
-
-        data = data.reset_index().rename(columns={'index': 'date'})
-        data['ticker'] = ticker
-        data['interval'] = interval.lower()
-
-        cols = list(data.columns)
-        cols.remove('ticker')
-        cols.remove('date')
-
-
-        data = data[['ticker', 'date'] + cols]
-
-        # now for pre and post items
-        thours = self._get_session_from_tperiod(meta)
-
-        data['date_local'] = data['date'].dt.tz_convert(thours['tz'])
-
-        data['regular_flag'] = 'N'
-        data['pre_flag'] = 'N'
-        data['post_flag'] = 'N'
-
-        data.loc[(data['date_local'].dt.time>=thours['regular_start']) &
-                (data['date_local'].dt.time < thours['regular_end']), 'regular_flag'] = 'Y'
-
-        data.loc[(data['date_local'].dt.time>=thours['pre_start']) &
-                (data['date_local'].dt.time < thours['pre_end']), 'pre_flag'] = 'Y'
-
-        data.loc[(data['date_local'].dt.time>=thours['post_start']) &
-                (data['date_local'].dt.time< thours['post_end']), 'post_flag'] = 'Y'
-
-        print(f'GOT for ticker')
-        return data# }}}
-
-    async def get_price_history(self, tickers, period=None, interval=None, start=None, end=None,# {{{
-            adjust=True, prepost=False):
-
-        if isinstance(tickers, str):
-            tickers = [tickers]
-
-        tasks = []
-        for t in tickers:
-            tmp = self._price_history(t, period=period, interval=interval, start=start,
-                    end=end, adjust=adjust, prepost=prepost)
-            tasks.append(tmp)
-        res = await asyncio.gather(*tasks)
-
-        return pd.concat(res)# }}}
-
     def _get_dates_from_period(self, period, asof=None):# {{{
+        """valid are 1d, 1mo, 1w, 1y, ytd, mtd, qtd"""
 
         if not asof:
             asof = pd.to_datetime('today')
@@ -521,7 +601,7 @@ class YFClient:
         elif asof.tzinfo != pytz.UTC:
             asof = asof.tz_convert(pytz.UTC)
 
-        if 'd' in period.lower():
+        if 'd' in period.lower() and period.lower() not in ['ytd', 'mtd', 'qtd']:
             start = asof - pd.Timedelta(days=int(period[:-1]))
 
         elif 'mo' in period.lower():
@@ -530,7 +610,7 @@ class YFClient:
         elif 'w' in period.lower():
             start = asof - relativedelta.relativedelta(weeks=int(period[:-1]))
 
-        elif 'y' in period.lower():
+        elif 'y' in period.lower() and period.lower() not in ['ytd']:
             start = asof - relativedelta.relativedelta(years=int(period[:-1]))
 
         elif period.lower() == 'ytd':
@@ -625,7 +705,11 @@ class YFClient:
             df.append({'symbol': symb, 'reco_symbol': reco_symbs,
                 'reco_score': reco_score})
 
-        return pd.concat([pd.DataFrame(x) for x in df])# }}}
+        if df:
+            return pd.concat([pd.DataFrame(x) for x in df])
+        else:
+            return pd.DataFrame(columns=['symbol', 'reco_symbol', 'reco_score'])
+        # }}}
 
     async def _get_markettime(self, region):# {{{
 
@@ -635,7 +719,7 @@ class YFClient:
         return await res.json()# }}}
 
     async def get_markettime(self, regions=['ar', 'au', 'br', 'ca', 'us', 'cn', 'de', 'es',#{{{
-        'fr', 'hk', 'in', 'it', 'jp', 'kr', 'mx', 'nz', 'sg', 'tw']):
+        'fr', 'hk', 'in', 'it', 'jp', 'kr', 'mx', 'nz', 'sg', 'tw', 'gb']):
 
         if isinstance(regions, str):
             regions = [regions]
@@ -679,6 +763,12 @@ class YFClient:
 
             ffres.append(tmp)
 
+        ffres = pd.concat(ffres, axis=1).T
+        ffres['open'] = pd.to_datetime(ffres['open'])
+        ffres['close'] = pd.to_datetime(ffres['close'])
+        ffres['open'] = ffres.apply(lambda x: x['open'].tz_convert(x['pytz_name']), axis=1)
+        ffres['close'] = ffres.apply(lambda x: x['close'].tz_convert(x['pytz_name']), axis=1)
+
         return pd.concat(ffres, axis=1).T# }}}
 
     async def _get_esg_score(self, symbol, count=5):# {{{
@@ -710,7 +800,10 @@ class YFClient:
         final = []
         for sym, res in fres.items():
 
-            list_res = res['esgPeerScores']['result'][0]['esgPeerScoresDocuments']
+            try:
+                list_res = res['esgPeerScores']['result'][0]['esgPeerScoresDocuments']
+            except:
+                continue
 
             for d in list_res:
                 tmp = {}
@@ -776,78 +869,13 @@ class YFClient:
 
         return pd.DataFrame(res)# }}}
 
+    def camel_to_snake(self, cols):# {{{
+        if isinstance(cols, str):
+            cols = [cols]
+        res = []
+        for c in cols:
+            res.append(self.snake_pattern.sub('_', str(c)).lower())
+        return res# }}}
 
 
-# implement search
-# yf = YFClient()
-# await yf.connect()
-
-
-# s = await yf.get_search(['BYND', 'Tesla'], stype='news', count=200)
-# 
-
-# await yf.disconnect()
-# 
-
-
-# esg = await yf.get_esg_peer_scores(['AAPL', 'TSLA'], count=20)
-
-# times = await yf.get_markettime()
-
-
-# trend = await yf.get_trending(regions=['us', 'hk'], count=10)
-
-# qs = await yf.get_quote(['AAPL', 'TSLA'])
-
-# data = await yf.get_all_equity_reference()
-
-# data = await yf.get_equity_reference(region='us', max_results=10)
-
-
-# data = await yf.get_price_history(['AAPL', 'TSLA'], period='1mo', end=pd.to_datetime('today',utc=True), interval='1d')
-
-
-# await yf.disconnect()
-# 
-# trend = yf.get_trending(regions=['us', 'gb'], count=10)
-
-
-# x = yf._price_history('AAPL', period='1d', interval='1m')
-
-# df= yf._price_history('GOOG', start=pd.to_datetime('2021-09-29 00:00:00', utc=True),
-#             end=pd.to_datetime('2021-09-29 11:59:59', utc=True), interval='5m', prepost=True)
-
-
-
-# start = pd.to_datetime('2021-09-01 00:00:00', utc=True)
-# end = pd.to_datetime('2021-09-30 23:00:00', utc=True)
-# 
-# df = yf._price_history('^GSPC', start=start, end=end, interval='5m', prepost=False)
-# 
-# df['hour'] = df['date_local'].dt.hour
-# df['minute'] = df['date_local'].dt.minute
-# df['time'] = df['date_local'].dt.time
-
-# x = yf.get_price_history('AZN.L', period='1mo', interval='1d', adjust=True)
-# x = yf.get_price_history('AZN.L', period='1mo', interval='1d', adjust=True)
-
-# x = yf.get_price_history('AAPL', start='20210505', end='20210505', interval='1d', adjust=True)
-
-# ref = yf.get_all_equity_reference()
-
-
-# quotes = yf._get_quote(test['symbol'].tolist())
-
-# qs = yf._get_quote_summary('GME')
-
-# q = yf.get_quote(['GBPUSD=X', 'EURUSD=X', 'AAPL'], ['bid', 'ask'])
-# # q = yf.get_quote(['BTC-USD'])
 # q['spread'] = ((q['ask'] - q['bid']) / ((q['ask'] + q['bid']) / 2)) * 10000
-# q
-
-
-
-# test = yf.get_equity_reference(region='us', max_results=100)
-
-# df = yf.get_equity_reference(region='us', max_results=None)
-# uk = yf.get_equity_reference(region='gb', max_results=None)
